# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# Inferable keras quantizer signature parameters:
NUM_BITS = 'num_bits'
SIGNED = 'signed'
THRESHOLD = 'threshold'
PER_CHANNEL = 'per_channel'
MIN_RANGE = 'min_range'
MAX_RANGE = 'max_range'
CHANNEL_AXIS = 'channel_axis'
INPUT_RANK = 'input_rank'
CLUSTER_CENTERS = 'cluster_centers'